
#include "reader.hpp"

using namespace std;
//using namespace Eigen;

int main(int argc, char* argv[])
{
   if (argc <= 3)
   {
      cout << "Not enough input arguments." << endl;
      return -1;
   }
   string data_filename = argv[1];
   string times_filename = argv[2];
   string out_filename = argv[3];

   unsigned int N = 32; //average estimate over N samples
   unsigned int dec = 4; //decimation factor

   vector<double> times = read_times(times_filename);
   vector<complex<double> > samples = read_samples(data_filename);

   double dt;  //time axis resolution
   double fs;  //sampling frequency
   double fs_decimated;

   if (times.size() >= 2)
   {
      dt = times[1] - times[0];
      fs = samples.size() / (times.size() * dt);
      cout << "Sampling frequency: " << fs << endl;
      fs_decimated = fs / dec;
      cout << "After decimation: " << fs_decimated << endl;
   }
   else
   {
      cout << "Time samples were not loaded correctly." << endl;
      return -1;
   }

   cout << "Samples: " << samples.size() << endl;
   cout << "Times: " << times.size() << endl;

   vector<double> estimate = freq_estimate(samples, N);
   vector<double> decimated = decimate(estimate, dec);

   FHRdata results = get_heart_rates(decimated, fs_decimated);

   write_file("bpms" + out_filename, results.bpms);
   write_file("cost" + out_filename, results.cost);

   return 0;
}

FHRdata get_heart_rates(const vector<double>& signal, const double& fs)
{
   double steps = 2; //calculations per second (based on ideal heart rate)
   bpm_params bpm_range = {140, 70, 250};
   double cycles_per_segment = 3;                                     
   
   unsigned int segment_size = round(cycles_per_segment * 60 * fs / bpm_range.start);
   cout << "Initial segment size: " << segment_size << endl;
   unsigned int stepsize = round(fs / steps);
   unsigned int previous_max = round(60 * fs / bpm_range.start);
   
   unsigned int pos = 0;
   vector<double> segment;
   vector<double> heart_rates;
   vector<double> cost;
   unsigned int tri_size = segment_size;

   while (pos + segment_size < signal.size())
   {
      //Get signal segment
      segment = vector<double>(signal.begin() + pos, signal.begin() + segment_size + pos);
      cout << "Segment size: " << segment.size() << endl;

      //Get maximum of autocorrelation function
      afvalues afval = get_af_params(segment, tri_size, previous_max, fs, bpm_range);

      //Calculate BPM
      double heart_rate = 60 * fs / afval.max;

      //Add BPM to list
      cost.push_back(afval.cost);
      heart_rates.push_back(heart_rate);

      //Increment position
      pos = pos + stepsize;

      //Save previous AF maximum
      previous_max = afval.max;

      //for adaptive segment length
      tri_size = segment_size;
      segment_size = round(cycles_per_segment * afval.max);
      cout << "Max element: " << afval.max << endl;
      cout << "Next segment size: " << segment_size << endl;
   }

   FHRdata data = {heart_rates, cost};
   return data;
}

vector<double> get_tri_window(const unsigned int tri_size, const unsigned int seg_size, const unsigned int maxelem)
{
   assert(tri_size & 1);
   vector<double> tri;
   tri.resize(seg_size, 0.0);
   cout << "Tri size: " << tri_size << endl;
   cout << "Seg size: " << seg_size << endl;
  
   if (tri_size == 0 || seg_size == 0 || maxelem == 0 || maxelem >= seg_size)
      return tri;

   unsigned int tri_half = (tri_size - 1) / 2;
   int start_pos = maxelem - tri_half;
   int last_pos = maxelem + tri_half;

   if (start_pos < 0)
      start_pos = 0;
   if (last_pos >= int(seg_size) - 1)
      last_pos = seg_size - 1;

   for (int i = start_pos; i <= last_pos; i++)
   {
      double val = 1.0 - std::fabs(double(i - 1 - int(maxelem))) / tri_half;
      tri[i] = val;
   }
   
   //cout << "\nTriangle window" << endl;
   //for (size_t i = 0; i < tri.size(); i++)
   //   cout << i << ": " << setprecision(8) << tri[i] << endl;
#if 0
   // Matlab method
   unsigned int tri_half = (size - 1) / 2;

   for (size_t i = 0; i < tri_half; i++)
   {
      double val = 2 * i / tri_length;
      tri[i] = val;
      *(tri.end() - 1 - i) = val;
   }
   tri[tri_half + 1] = 1.0;

   /* Heavier dampening for low lag values
   for (size_t i = 0; i < maxelem; i++)
      tri[i] = static_cast<double>(i) / maxelem;

   tri[maxelem] = 1.0;

   for (size_t i = size - 1; i > maxelem; i--)
      tri[i] = 1.0 - (static_cast<double>(i) - maxelem) / (size - maxelem);
   */
#endif
   return tri;
}

afvalues get_af_params(const vector<double>& signal, const unsigned int tri_size, const unsigned int previous_max, const double fs, const bpm_params& bpm_range)
{
   unsigned int minlag = round(60 * fs / bpm_range.max);
   unsigned int maxlag = round(60 * fs / bpm_range.min);
   maxlag = min(maxlag, static_cast<unsigned int>(signal.size()));
   cout << "maxlag: " << maxlag << endl;
   
   //Compute AF
   vector<double> af_full = crosscorr(signal, maxlag);

   cout << "Segment size: " << signal.size() << endl;
   cout << "AF size: " << af_full.size() << endl;

   //Default AF parameters
   afvalues afv = {static_cast<unsigned int>(60 * fs / bpm_range.max), 1.0};

   //for static segment sizes
   //unsigned int tri_size = signal.size() + (signal.size() % 2 == 0);   //odd length
   
   unsigned int tri_size_odd = tri_size + (tri_size % 2 == 0);
   vector<double> window = get_tri_window(tri_size_odd, signal.size(), previous_max);

   //window size should match AF
   window.resize(maxlag, 0.0);

   //multiply AF with window
   vector<double> af_windowed;
   af_windowed.resize(maxlag, 0.0);
   
   cout << "Window size: " << window.size() << endl;

   assert(af_windowed.size() == af_full.size());
   assert(af_windowed.size() == window.size());

   for (size_t i = minlag; i < maxlag; i++)
      af_windowed[i - minlag] = af_full[i] * window[i];
   
   cout << "AF size: " << af_full.size() << ", window size: " << window.size() << endl;

   vector<double>::iterator max_iter = max_element(af_windowed.begin(), af_windowed.end());
   unsigned int max_elem = max_iter - af_windowed.begin();

   if (af_full[0] == 0)
      return afv;
   
   double energy = af_full[0];
   cout << "Signal energy: " << setprecision(8) << energy << endl;
   double cost = 1 - af_windowed[max_elem] / energy;
   cout << "Max value: " << setprecision(8) << af_windowed[max_elem] << endl;

   afv.max = max_elem + minlag + 2; //why + 2? check against Matlab
   afv.cost = cost;
   
   return afv;
}

vector<double> crosscorr(const vector<double>& signal, size_t maxlag)
{
   vector<double> ccfunction;
   ccfunction.resize(maxlag);
   
   for (size_t l = 0; l < maxlag; l++)
   {
      double corrsum = 0.0;
      for (size_t i = 0; i < signal.size() - l; i++)
      {
         double corr = signal[i] * signal[i + l];
         corrsum += corr;
      }
      ccfunction[l] = corrsum;
   }

   return ccfunction;
}

vector<double> decimate(const vector<double>& signal, const unsigned int& dec)
{
   vector<double> decimated;

   if (dec == 0) 
      return decimated;

   cout << "Signal size: " << signal.size() << endl;
   cout << "Samples after decimation: " << floor(signal.size() / dec) << endl;

   //make integer division round upwards
   unsigned int sample_count = (signal.size() + dec - 1) / dec;
   decimated.resize(sample_count);

   unsigned int count = 0;
   for (size_t i = 0; i < signal.size(); i += dec)
   {
      decimated[count] = signal[i];
      count++;
   }

   return decimated;
}

vector<double> freq_estimate(const vector<complex<double> >& signal, const unsigned int& N)
{
   vector<double> estimate;

   // use N + 1 for 1-element segment overlap
   if (signal.size() < (N + 1))
      return estimate;

   estimate.resize(signal.size() - (N + 1));

   vector<complex<double> > segment;

   //cout << "\nFrequency estimate:" << endl;
   for (size_t i = 0; i < signal.size() - (N + 1); i++)
   {
      //initialize or update segment
      segment = vector<complex<double> >(signal.begin() + i, signal.begin() + N + i + 2);

      vector<complex<double> > freqs;
      freqs.resize(segment.size() - 1, 0.0);

      //loop through segment
      for (size_t j = 0; j < segment.size() - 1; j++)
         freqs[j] = segment[j + 1] * conj(segment[j]);

      //compute mean
      if (freqs.size() == 0)
         return estimate;

      complex<double> freqsum(0.0);
      for (size_t k = 0; k < freqs.size(); k++)
         freqsum += freqs[k];

      //cout << "freqsum: " << freqsum << endl;
      complex<double> meanfreq = freqsum / (double) freqs.size();
      //cout << "meanfreq: " << meanfreq << endl;
      double meanangle = arg(meanfreq);
      //cout << "meanangle: " << meanangle << endl;

      //add result to estimate vector
      estimate[i] = meanangle;
   }
   return estimate;
}

vector<double> read_times(const string& filename)
{
   vector<double> data;   
   char buffer[256];

   ifstream in_file;
   in_file.open(filename.c_str());

   in_file.getline(buffer, 256);   
   while(in_file.good())
   {
      double val = strtod(buffer, NULL);
      data.push_back(val);
      in_file.getline(buffer, 256);
   }

   in_file.close();

   return data;
}

vector<complex<double> > read_samples(const string& filename)
{
   vector<complex<double> > data;   
   char buffer[256];

   ifstream in_file;
   in_file.open(filename.c_str());

   in_file.getline(buffer, 256);   
   while(in_file.good())
   {
      char* endp;
      double real = strtod(buffer, &endp);
      double imag = strtod(endp, NULL);
      complex<double> val = complex<double>(real, imag);
      data.push_back(val);
      in_file.getline(buffer, 256);
   }

   in_file.close();

   return data;
}

void write_file(const string& filename, const vector<double>& data)
{
   ofstream out_file;
   out_file.open(filename.c_str());

   for (size_t i = 0; i < data.size(); i++)
      out_file << data.at(i) << endl;

   out_file.close();
}
