close all;
clear all;

%% Algorithm parameters

% Clutter filter
params.highpass = 1;           % Clutter filter instead of lowpass
params.chebystop = 0.35;       % End of stopband, clutter filter
params.chebypass = 0.45;       % Start of passband, clutter filter

% Frequency estimate
params.N = 32;                 % Estimate averaging length
params.pthreshold = 75;        % Power threshold in dB
params.envelope = 0;           % Use envelope of signal, no freq.est
                               %    (use with highpass = 0)
% Decimation
params.antialias = 0;          % Enable anti-aliasing
params.dec = 4;                % Decimation factor, set = 1 to disable

% Autocorrelation
params.bpm_min = 70;           % Minimum measurable BPM
params.bpm_max = 250;          % Maximum measurable BPM
params.cycles = 2;             % Cycles (of BPM_min) per segment 
params.steps = 2;              % Calculations per second
params.triwindow = 0;          % Triangular windowing of AF

% Plotting parameters
params.fullsets = 0;           % Plot full datasets
params.calcavg = 0;            % Average calculations over 3
params.costthreshold = 0.8;    % Cost threshold before rejection
params.stable = 1;             % # of stable values after rejection
params.plotbpms = 1;           % Plot results (w/ spectrum)

% Save plots
params.savefinal = 0;          % Save plots with final results
params.saveaf = 0;             % Save AF and segment data

manualcomp = 1;                % Calculate difference between manual and auto

%% Viable datasets and segments

dataset = [2 3 4 6 7 6 10];
dataname = ['A 2'; 'A 3'; 'A 4'; 'A 6'; 'A 7'; 'B 6'; 'B10'];
setnums = [1 1 1 1 1 2 2];
setxlims = [5.5 9.5; 7.0 11.0; 2.0 10.0; 3.0 10.5; 10.0 16.0; 38.0 52.0; 22.5 34.5];
setylims = [125 145; 130 150; 135 155; 125 145; 60 160; 125 145; 135 155];
% RMSE intervals
%setxlims = [5.5 9.5; 7.4 10.5; 1.7 6.0; 4.8 7.3; 12.2 14.8; 43.9 52.0; 23.2 33.8];

% Intervals where signal is stable enough that MSE is meaningful
maninterv = [5.5 9.5; 7.4 10.5; 1.7 6.0; 4.8 7.3; 12.5 15.0; 43.9 52.0; 23.2 33.8];

pstart = 1;
pend = length(dataset);

% dataset A7
%pstart = 5;
%pend = 5;

% dataset B6
%pstart = 6;
%pend = 6;

% dataset B10
%pstart = 7;
%pend = 7;

% dataset A4
%pstart = 3;
%pend = 3;

% Plot appearance
labelsize = 13;
axissize = 12;
titlesize = 14;

for p = pstart : pend
    setnum = int2str(dataset(p));
    if (setnums(p) == 2)
        zstr = '00';
    else 
        zstr = '0';
    end
    
    fprintf('\nDataset %s\n', setnum);
    
    eval(['load data/PWDoppler', zstr, setnum]);
    times = DopplerData.times;
    doppler = DopplerData.data;
    
    setstr = [zstr, setnum];
    
    %fs = length(doppler) / (times(end) - times(1));
    % Use known fs
    if (strcmp(setstr, '006') || strcmp(setstr, '0010'))
        fs = 2572;
    else
        fs = 2550;
    end
    
    [spectrum, bpms, afcost, estimate] = Umojatest(DopplerData, params, setstr, setxlims(p, :));
    
    eval(['save(''bpms', zstr, setnum, '.mat'', ''bpms'');']); 
    
    name = ['Dataset ', dataname(p, :), ', frequency spectrum'];
    
    % Averaging of automatic measurements
    if (params.calcavg)
        bpms_avg = bpms(:, 3);
        ht_avg = bpms(:, 2);
        for i = 2 : length(bpms(:, 1)) - 1
            bpms_avg(i) = mean(bpms(i - 1 : i + 1, 3));
            ht_avg(i) = mean(bpms(i - 1 : i + 1, 2));
        end
        bpms(:, 3) = bpms_avg;
        bpms(:, 2) = ht_avg;
    end
    
    % Load manual measurements
    eval(['load manual/manual', zstr, setnum]);
    
    if (params.plotbpms)
        
        h = figure(p);
        
        % Plot Doppler spectrum
        subplot(3, 1, 1)
        if (params.highpass)
            gain = -90;
            dyn = 60;
        else
            gain = -100;
            dyn = 50;
        end
        spectrum_log = 256 * (gain + 20 * log10(abs(spectrum))) / dyn;
        image([times(1) times(end)], [-0.5 * fs 0.5 * fs], spectrum_log)
        colormap(gray(256))
        set(gca, 'YDir', 'reverse')
        if (params.fullsets == 0)
            xlim([setxlims(p, 1) setxlims(p, 2)])
        end
        if (~params.highpass)
            ylim([-0.15 * fs 0.15 * fs])
        end
        title(name, 'fontSize', titlesize);
        xlabel('time [s]', 'fontSize', labelsize)
        ylabel('frequency [Hz]', 'fontSize', labelsize)
        set(gca, 'fontSize', axissize)
        
        % Plot mean angular frequency estimate
        subplot(3, 1, 2)
        grid on;
        plot(linspace(times(1), times(end), length(estimate)), estimate, 'r-')
        if (params.fullsets == 0)
            xlim([setxlims(p, 1) setxlims(p, 2)])
        end
        if (~params.envelope)
            if (params.highpass)
                ylim([-pi pi])
            else
                ylim([-0.3 0.3])
            end
        end
        xlabel('time [s]', 'fontSize', labelsize)
        ylabel('angle [rad]', 'fontSize', labelsize)
        title('Mean angular frequency estimate', 'fontSize', titlesize)
        set(gca, 'fontSize', axissize)
        
        % Plot calculated FHR
        subplot(3, 1, 3)
        hold on
        grid on
        
        % Manual measurements color
        mcolor = [0 0 1];

        linerate = length(spectrum) / (times(end) - times(1));
        
        munfcolor = [1 0 1];
        plot(bpms_manual(:, 1), min(bpms_manual(:, 3), 250), '-b')
        scatter(bpms_manual(:, 1), min(bpms_manual(:, 3), 250), 50, mcolor, 'filled')

        error = afcost;
        ref = 140 * ones(1, length(bpms(:, 1)));
        % 140 BPM reference line
        plot(bpms(:, 1), ref, 'b-')
        % Cost function
        %eref = (20 * threshold + 140) * ones(1, length(axis));
        %plot(axis,  eref, 'm-')
        %plot(axis,  20 * error + 140, 'r-')

        prevaccepted = 0;
        colors = ones(length(error), 3);
        for n = 1 : length(error)
            if (error(n) < params.costthreshold)
                if (prevaccepted >= params.stable)
                    colors(n, :) = [0 0.8 0];
                else
                    colors(n, :) = [0.8 0.8 0];
                end
                prevaccepted = prevaccepted + 1;
            else
                colors(n, :) = [1 0 0];
                %reset counter
                prevaccepted = 0;
            end
        end
        
        %plot(axis, bpms, 'k-')
        %scatter(axis, bpms, 50, colors, 'filled')
        
        %plot(bpms(:, 1), bpms(:, 3), 'k-')
        %scatter(bpms(:, 1), bpms(:, 3), 50, colors, 'filled')
        
        plot(bpms(:, 1), bpms(:, 3), 'k-')
        scatter(bpms(:, 1), bpms(:, 3), 50, colors, 'filled')
        
        %scatter(axis, bpms, 50, 'k', 'filled')
        %scatter(axis, bpms_avg, 50, colors, 'filled')

        set(gcf,'Renderer','zbuffer')
        
        if (params.fullsets == 1)
            xlim([DopplerData.times(1) DopplerData.times(end)])
        else
            xlim(setxlims(p, :))
        end
        
        ylim(setylims(p, :))
        ylabel('FHR [BPM]', 'fontSize', labelsize)
        xlabel('time [s]', 'fontSize', labelsize)
        title('Fetal heart rate, Matlab vs. manual', 'fontSize', titlesize)
        hold off
        
        set(gca, 'fontSize', axissize)
        config = [num2str(params.pthreshold), 'dB-', num2str(params.dec), 'dec-', num2str(params.cycles), 'c-tri', num2str(params.triwindow), '-aa', num2str(params.antialias), '-N', num2str(params.N)];
        set(h,'Name',['Set ', zstr, setnum, ', ', num2str(params.pthreshold), ' dB, ', num2str(params.dec), ' dec, ', num2str(params.cycles), ' cycles @ ', num2str(params.bpm_min), ' BPM, aa ', num2str(params.antialias), ' N', num2str(params.N)],'NumberTitle','off')
        if (params.savefinal)
            saveas(h, ['fig/bpms-set', zstr, setnum, '-', config], 'fig');
        end
    end
    
    if (manualcomp)
        fprintf('Segment duration: %f s\n', maninterv(p, 2) - maninterv(p, 1));
        maninterp = [];
        % Use cycle period (:, 2) instead of the non-linear BPM (:, 3) 
        maninterp = interp1(bpms_manual(:, 1), bpms_manual(:, 2), bpms(:, 1), 'linear');
        maninterpbpm = interp1(bpms_manual(:, 1), bpms_manual(:, 3), bpms(:, 1), 'linear');

        count = 0;
        acount = 0;
        for i = 1 : length(maninterp)
            if (~isnan(maninterp(i)) && bpms(i, 1) > maninterv(p, 1) && bpms(i, 1) < maninterv(p, 2))
                count = count + 1;
                manualms(count) = maninterp(i);
                manualbpm(count) = maninterpbpm(i);
                
                automs(count) = bpms(i, 2);
                autobpm(count) = bpms(i, 3);
            end
        end
        
        % Error (s)
        manerror = automs - manualms;
        manerror_mean = mean(manerror);
        manerror_std = std(manerror);
        % Error (BPM)
        manerrorbpm = autobpm - manualbpm;
        manerrorbpm_mean = mean(manerrorbpm);
        manerrorbpm_std = std(manerrorbpm);
        
        % Bias
        biasms = mean(automs) - mean(manualms);
        biasbpm = mean(autobpm) - mean(manualbpm);

        % Mean Square Error
        %mse = mean((automs - manualms) .^ 2);
        %msebpm = mean((autobpm - manualbpm) .^ 2);
        % Root Mean Square Error
        %rmse = sqrt(mse);
        %rmsebpm = sqrt(msebpm);
        
        % time
        %fprintf('Error: %f +/- %f ms\n', manerror_mean * 1000, manerror_std * 1000);
        %fprintf('Bias: %f ms\n', biasms * 1000);
        %fprintf('RMSE: %f ms\n\n', rmse * 1000);
        
        % heart rate
        fprintf('Error: %f +/- %f BPM\n', manerrorbpm_mean, manerrorbpm_std);
        fprintf('Bias: %f BPM\n', biasbpm);
        %fprintf('RMSE: %f BPM\n', rmsebpm);
        
        %fprintf('Latex: SN & %1.2f $\\pm$ %1.2f & %1.2f \\\\ \n', manerrorbpm_mean, manerrorbpm_std, biasbpm);
    end
    
    %{
    %FHRV = abs(bpms(2 : end, 2) - bpms(1 : end - 1, 2));
    %fprintf('mean FHRV: %f ms\n', mean(FHRV) * 1000);
    
    % Find resolution of measurements
    bpmdiff = abs(bpms(2 : end, 3) - bpms(1 : end - 1, 3));
    figure(222 + p)
    subplot(2, 1, 1)
    grid on
    plot(bpms(2 : end, 1), bpmdiff)
    xlabel('time [s]')
    ylabel('heart rate [BPM]')
    ylim([0, 1])
    subplot(2, 1, 2)
    grid on
    plot(bpms(2 : end, 1), FHRV * 1000)
    xlabel('time [s]')
    ylim([0, 10])
    ylabel('heart rate variability [ms]')  
    %}
    
    %{
    % Plot cost function
    figure(200 + p)
    subplot(2, 1, 1)
    hold on
    grid on
    plot(bpms_manual(:, 1), min(bpms_manual(:, 3), 250), '-b')
    scatter(bpms_manual(:, 1), min(bpms_manual(:, 3), 250), 50, mcolor, 'filled')
    plot(bpms(:, 1), ref, 'b-')
    plot(bpms(:, 1), bpms(:, 3), 'k-')
    scatter(bpms(:, 1), bpms(:, 3), 50, colors, 'filled')
    hold off
    ylabel('FHR [BPM]', 'fontSize', labelsize)
    xlabel('time [s]', 'fontSize', labelsize)
    title('Fetal heart rate, Matlab vs. manual', 'fontSize', titlesize)
    if (dataset(p) == 7)
        ylim([60 170])
    else
        ylim([120 170])
    end
    if (params.fullsets == 1)
        xlim([DopplerData.times(1) DopplerData.times(end)])
    else
        xlim(setxlims(p, :))
    end
    % Cost function
    subplot(2, 1, 2)
    hold on
    eref = threshold * ones(1, length(bpms(:, 1)));
    plot(bpms(:, 1), eref, 'm-')
    plot(bpms(:, 1), error, 'r-')
    grid on
    hold off
    set(gca, 'fontSize', axissize)
    ylabel('cost', 'fontSize', labelsize)
    xlabel('time [s]', 'fontSize', labelsize)
    title('Cost function', 'fontSize', titlesize)
    ylim([0 1])
    if (params.fullsets == 1)
        xlim([DopplerData.times(1) DopplerData.times(end)])
    else
        xlim(setxlims(p, :))
    end
    %}
    
    %{
    % find specific segments
    figure(60 + p)
    hold on
    stem(bpms(:, 1), bpms(:, 3), 'k-')
    plot(bpms_manual(:, 1), min(bpms_manual(:, 3), 250), '-b')
    scatter(bpms_manual(:, 1), min(bpms_manual(:, 3), 250), 50, mcolor, 'filled')
    plot(bpms(:, 1), bpms(:, 3), 'k-')
    scatter(bpms(:, 1), bpms(:, 3), 50, colors, 'filled')    
    scatter(bpms(:, 1), linspace(1, length(bpms(:, 3)), length(bpms(:, 3))), 50, 'k', 'filled')
    hold off
    %}
end